/*
 * mn.c
 *
 *  Created on: Jan 23, 2016
 *      Author: vderma
 */


// ****************************************************************
#include <stdio.h>
#include <stdlib.h>

// ****************************************************************
// Position of the decimal point
#define ACC 16



#define NB 5
#define NA 5
#define NX 8



// ****************************************************************
typedef long int Q32FIXED ;



// ****************************************************************
float b[NB] = { -0.1, 0.2, -0.5,  0.2, 0.1 } ;
float a[NA] = { 1.0, 0.3,-0.2, 0.5,  -0.1} ;
float x[NX] = {1.1, 2.2, 0.1, 0.4, 1.1, 2.2, 1.0, 0.5 } ;
float y[NX] ;

Q32FIXED bQ32[NB] ;
Q32FIXED aQ32[NA] ;
Q32FIXED xQ32[NX] ;
Q32FIXED yQ32[NX] ;




// ****************************************************************
Q32FIXED float2Q32( float a )
{
	Q32FIXED res ;
	return res =  a* ((float)(1<<ACC)) ;
}


// ****************************************************************
float Q322float( Q32FIXED a )
{
	float res = (1<<ACC) ;
	return (float) a / res ;
}


// ****************************************************************
inline Q32FIXED addQ32( Q32FIXED a, Q32FIXED b )
{
	return a+b ;
}

// ****************************************************************
inline Q32FIXED subQ32( Q32FIXED a, Q32FIXED b )
{
	return a-b ;
}

// ****************************************************************
inline Q32FIXED mulQ32( Q32FIXED a, Q32FIXED b )
{
	return (a*b)>>ACC ;
}

// ****************************************************************
inline Q32FIXED divQ32( Q32FIXED a, Q32FIXED b )
{
	return ((a<<ACC)/b) ;
}



// ****************************************************************
// FIR filtering using floating point arithmetic
//  b[] filter coefficients
//  nb number of coefficients in array b[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void FIR( float b[], int nb, float x[], int nx, float  y[] )
{
	int i, j, it ;
	float s ;
	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j]*b[j] ;
		y[i] = s ;
	}
}


// ****************************************************************
// IIR filtering using floating point arithmetic
//  b[], a[] filter coefficients
//  nb number of coefficients in array b[]
//  na number of coefficients in array a[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void IIR( float b[], int nb, float a[], int na, float x[], int nx, float  y[] )
{
	int i, j, it ;
	float s ;
	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j]*b[j] ;
		for( j = 1, it = (i<na) ? i:na ; j <= it ; j++ )
			s += y[i-j]*a[j] ;
		y[i] = s ;
	}
}



// ****************************************************************
// FIR filtering using fixed point arithmetic
//  b[] filter coefficients
//  nb number of coefficients in array b[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void FIRQ32( Q32FIXED b[], int nb, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;
	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s = addQ32(s,mulQ32( x[i-j], b[j] )) ;
		y[i] = s ;
	}
}

// ****************************************************************
// fastFIR filtering using fixed point arithmetic
//  b[] filter coefficients
//  nb number of coefficients in array b[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void fastFIRQ32( Q32FIXED b[], int nb, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;

	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j]*b[j] ;
		y[i] = s>>ACC ;
	}
}

// ****************************************************************
// superfastFIR filtering using fixed point arithmetic
//  b[] filter coefficients
//  nb number of coefficients in array b[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void superfastFIRQ32( int b[], int nb, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;

	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j]*b[j] ;
		y[i] = s ;
	}
}

// ****************************************************************
// IIR filtering using fixed point arithmetic
//  b[], a[] filter coefficients
//  nb number of coefficients in array b[]
//  na number of coefficients in array a[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void IIRQ32( Q32FIXED b[], int nb, Q32FIXED a[], int na, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;

	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s = addQ32(s,mulQ32( x[i-j], b[j] )) ;
		for( j = 1, it = (i<na) ? i:na ; j <= it ; j++ )
			s = addQ32(s,mulQ32( y[i-j], a[j] )) ;
		y[i] = s ;
	}
}

// ****************************************************************
// fastIIR filtering using fixed point arithmetic
//  b[], a[] filter coefficients
//  nb number of coefficients in array b[]
//  na number of coefficients in array a[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void fastIIRQ32( Q32FIXED b[], int nb, Q32FIXED a[], int na, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;

	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j] * b[j] ;
		for( j = 1, it = (i<na) ? i:na ; j <= it ; j++ )
			s += y[i-j] * a[j] ;
		y[i] = s>>ACC ;
	}
}


// ****************************************************************
// superfastIIR filtering using fixed point arithmetic
//  b[], a[] filter coefficients
//  nb number of coefficients in array b[]
//  na number of coefficients in array a[]
//  x[] signal of size nx
//  y[] signal output
// ****************************************************************
void superfastIIRQ32( int b[], int nb, int a[], int na, Q32FIXED x[], int nx, Q32FIXED  y[] )
{
	int i, j, it ;
	Q32FIXED s ;

	for( i = 0 ; i < nx ; i++ )
	{
		for( j = s = 0, it = (i<nb) ? i:nb ; j <= it ; j++ )
			s += x[i-j] * b[j] ;
		for( j = 1, it = (i<na) ? i:na ; j <= it ; j++ )
			s += y[i-j] * a[j] ;
		y[i] = s ;
	}
}

// ****************************************************************
void exampleFIRfiltering( void )
{
	int i ;

	// floating point FIR filtering ...
	FIR( b, NB, x, NX, y ) ;

	// Conversion float coefficients to fixed point ...
	for( i = 0 ; i < NB ; i++ )
		bQ32[i] = float2Q32(b[i]) ;
	for( i = 0 ; i < NX ; i++ )
		xQ32[i] = float2Q32(x[i]) ;

	// fixed point FIR filtering ...
	FIRQ32( bQ32, NB, xQ32, NX, yQ32 ) ;

	// Display the results
	printf( "\nDifferences between FIR filtering using floating and fixed point arithmetic\n");
	printf( "Output of the floating point FIR filter:\n");
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", y[i]) ;
	putchar( '\n') ;
	printf( "Output of the fixed point FIR filter:\n");
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", Q322float(yQ32[i])) ;
	putchar( '\n') ;

	printf( "Fast fixed point FIR filtering ...\n" ) ;
	fastFIRQ32( bQ32, NB, xQ32, NX, yQ32 ) ;
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", Q322float(yQ32[i])) ;
	putchar( '\n') ;

}

// ****************************************************************
void exampleIIRfiltering( void )
{
	int i;

	// floating point IIR filtering ...
	IIR( b, NB, a, NA, x, NX, y ) ;

	for( i = 0 ; i < NB ; i++ )
		bQ32[i] = float2Q32(b[i]) ;

	for( i = 0 ; i < NA ; i++ )
		aQ32[i] = float2Q32(a[i]) ;

	for( i = 0 ; i < NX ; i++ )
		xQ32[i] = float2Q32(x[i]) ;

	// fixed point FIR filtering ...
	IIRQ32( bQ32, NB, aQ32, NA, xQ32, NX, yQ32 ) ;

	// Display the results
	printf( "\nDifferences between IIR filtering using floating and fixed point arithmetic\n");
	printf( "Output of the floating point IIR filter:\n");
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", y[i]) ;
	putchar( '\n') ;
	printf( "Output of the fixed point IIR filter:\n");
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", Q322float(yQ32[i])) ;
	putchar( '\n') ;

	printf( "Fast fixed point FIR filtering ...\n" ) ;
	fastIIRQ32( bQ32, NB, aQ32, NA, xQ32, NX, yQ32 ) ;
	for(i = 0 ; i < NX ; i++ )
		printf( " %10.6f", Q322float(yQ32[i])) ;
	putchar( '\n') ;

}



// ****************************************************************
void exampleSingleOperations( void )
{
	Q32FIXED Q,Q1,Qr ;
	float F1 = 2.2, F = 1.1 ;

	printf( "Floating Point operations: %f+%f=%f\n", F,F1,F+F1) ;
	Q = float2Q32( F ) ;
	Q1 = float2Q32( F1 ) ;
	Qr = addQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f+%f=%f\n\n", Q322float(Q),Q322float(Q1),Q322float(Qr)) ;

	printf( "Floating Point operations: %f-%f=%f\n", F,F1,F-F1) ;
	Qr = subQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f-%f=%f\n\n", Q322float(Q),Q322float(Q1),Q322float(Qr)) ;

	printf( "Floating Point operations: %f*%f=%f\n", F,F1,F*F1) ;
	Qr = mulQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f*%f=%f\n\n", Q322float(Q),Q322float(Q1),Q322float(Qr)) ;

	printf( "Floating Point operations: %f+%f=%f\n", -3.0F,-2.5F,(-3.0F)+(-2.5F)) ;
	Q = float2Q32( -3.0F ) ;
	Q1 = float2Q32( -2.5F ) ;
	Qr = addQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f+%f=%f\n\n", Q322float(Q),Q322float(Q1),Q322float(Qr)) ;
	printf( "Floating Point operations: %f*%f=%f\n", -3.0F,-2.5F,(-3.0F)*(-2.5F)) ;
	Qr = mulQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f*%f=%f\n\n", Q322float(Q),Q322float(Q1),Q322float(Qr)) ;

	printf( "Floating Point operations: %f/%f=%f\n", -3.0F,-2.5F,(-3.0F)/(-2.5F)) ;
	Qr = divQ32(Q,Q1) ;
	printf( "Fixed Point operations: %f/%f=%f\n\n", Q322float(Q),Q322float(Q1), Q322float(Qr)) ;
}


// ****************************************************************
int main()
{
	exampleSingleOperations() ;
	exampleFIRfiltering() ;
	exampleIIRfiltering() ;


	return 0 ;
}
